"""
Handels seq http call
"""
from typing import List
import logging
import datetime

import requests

from . import utils, config

LOG = logging.getLogger(config.LOGGER_NAME)


class SeqSender:
    """
    Handels seq http call
    """

    @staticmethod
    def send(batch_objects: List[str]):
        """Sends and setups call to seq

        Parameters
        ----------
        batch_objects : List[str]
            seq log data
        """

        message_body = str.join("\n", batch_objects)
        headers = {
            "X-Seq-ApiKey": config.SEQ_APIKEY,
            "Content-type": "application/vnd.serilog.clef",
        }
        server_url = utils.url_add_trailing_slash(config.SEQ_SERVER)

        ok_result = SeqSender._send_call(server_url, headers, message_body)
        if not ok_result:
            SeqSender._fallback(message_body)

    @staticmethod
    def _send_call(server_url: str, headers: dict, message_body: str) -> bool:
        """Does http call to seq

        Parameters
        ----------
        server_url : str
            seq server
        headers : dict
            http call headers with api key
        message_body : str
            data to send

        Returns
        -------
        bool
            ok status
        """
        ok_status = True
        try:
            resonse = requests.post(
                f"{server_url}api/events/raw", data=message_body, headers=headers
            )
            resonse.raise_for_status()
        except Exception:
            LOG.exception("SeqPyLogger sending raised exception")
            ok_status = False
        return ok_status

    @staticmethod
    def _fallback(message_body: str):
        """Handels fallback when sending failed

        Parameters
        ----------
        message_body : str
            data meant to be send to seq
        """
        # Complete fallback
        with open("seqpylogger_error.log", "a", encoding="utf-8") as fstream:
            fstream.write(
                datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f%Z") + "\n"
            )
            fstream.write("Seq sending had an unhandeled exception\n")
            fstream.write(f"Failed to send the following log data:\n{message_body}\n")
        LOG.error(
            "SeqPyLogger failed to send messages to seq, see seqpylogger_error.log"
        )



"""
Handles incomming log records with buffer
"""
import datetime
import json
import logging
import traceback
from typing import Optional, Dict
from logging.handlers import BufferingHandler
from . import config, seqsender


LOG = logging.getLogger(config.LOGGER_NAME)


class SeqPyLoggerHandler(BufferingHandler):
    """
    Handles incomming log records with buffer
    """

    def __init__(self, capacity=10, formatter_style="%"):
        self.formatter_style = formatter_style
        super().__init__(capacity=capacity)

    def flush(self):
        try:
            self.acquire()
            if len(self.buffer) > 0:
                self.send_to_seq()
                super().flush()
        finally:
            self.release()

    def send_to_seq(self):
        """Prepares record for sending to seq"""
        batch_objects = []
        for record in self.buffer:
            # format_message monkey patchs LogRecord with .message
            record_args = SeqPyLoggerHandler.format_message(
                record, self.formatter_style
            )
            record_object = SeqPyLoggerHandler.format_record_for_seq(record)
            record_object.update(record_args)

            ex = SeqPyLoggerHandler.add_exception(record)
            if ex is not None:
                record_object.update({"@x": ex})

            batch_objects.append(json.dumps(record_object, separators=(",", ":")))

        seqsender.SeqSender.send(batch_objects)

    @staticmethod
    def format_record_for_seq(record: logging.LogRecord) -> dict:
        """Creates seq record object

        Parameters
        ----------
        record : logging.LogRecord
            Record to handle

        Returns
        -------
        dict
            formatted according to seq specs
        """
        record_object = {
            "@t": datetime.datetime.fromtimestamp(record.created).strftime(
                "%Y-%m-%dT%H:%M:%S.%f%Z"
            ),
            "@l": record.levelname,
            "@mt": record.msg,
            "@m": record.message,
            "@@Logger": record.name,
            "@@Path": record.pathname,
            "@@Line": record.lineno,
            "@@Function": record.funcName,
            "@@Thread": record.threadName,
            "@@Pid": record.process,
            "@@Environment": config.ENVIRONMENT,
        }

        return record_object

    @staticmethod
    def add_exception(record: logging.LogRecord) -> Optional[str]:
        """Addeds traceback data to log message

        Parameters
        ----------
        record : logging.LogRecord
            LogMessage to handle

        Returns
        -------
        Optional[str]
            Returns message with added traceback or None
        """
        if record.exc_info and any(record.exc_info):
            return str.join("", traceback.format_exception(*record.exc_info))
        return None

    @staticmethod
    def format_message(
        record: logging.LogRecord, formatter_style: str
    ) -> Dict[str, str]:
        """Prepares message shape for sending to seq as this requires a different syntax

        Parameters
        ----------
        record : logging.LogRecord
            LogMessage to handle

        formatter_style : str
            Log message formatting style should be "%"

        Returns
        -------
        dict
            dict of extra arguments in message
        """
        record_args: Dict[str, str] = {}
        if formatter_style != "%":
            logging.warning("SeqPyLogger Unimplemented formatting style")
            return record_args

        # Prevent logging not str type
        if not isinstance(record.msg, str):
            record.msg = str(record.msg)

        try:
            record.message = record.msg % record.args
        except TypeError:
            LOG.warning("SeqPyLogger message formatting failed - (%s)", record.msg)
            record.message = record.msg
        
        # Fixes %d not being replaced
        record.msg = record.msg.replace("%d", "%s")

        for i, arg in enumerate(record.args):
            record_args.update({"arg_%d" % i: str(arg)})
            record.msg = record.msg.replace("%s", "{arg_%d}" % i, 1)

        return record_args



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
from os import path
import pathlib

VERSION = "1.0.11"

this_directory = path.abspath(path.dirname(__file__))
long_description = pathlib.Path(this_directory, "README.md").read_text()

setup(
    name="seqpylogger",
    version=VERSION,
    description="python loghandler for seq",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Triple",
    packages=["seqpylogger"],
    install_requires=["requests"],
    include_package_data=True,
    url='https://github.com/wearetriple/seqpylogger',
    download_url=f"https://pypi.io/packages/source/s/seqpylogger/seqpylogger-{VERSION}.tar.gz",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: System :: Logging',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3'
    ]
)



"""
Main module for referencing handler
"""

import atexit
import time
import queue
import threading
from logging import getLogger, NullHandler, LogRecord
from logging.handlers import QueueHandler
from .seqpyloggerhandler import SeqPyLoggerHandler
from .seqpyloggerqueuelistener import SeqPyLoggerQueueListener
from .utils import set_interval
from . import config

LOCK = threading.Lock()


class SeqPyLogger(QueueHandler):
    """LogHandler for seq"""

    def __init__(self, buffer_capacity: int = 10, disable_on_exit_flush=False):
        self.log_queue: queue.Queue = queue.Queue(-1)
        self.log_handler = SeqPyLoggerHandler(capacity=buffer_capacity)

        self.queue_listener = SeqPyLoggerQueueListener(self.log_queue, self.log_handler)

        self.queue_listener.start()

        self.timout_flush()

        if not disable_on_exit_flush:
            atexit.register(self.__exit_flush)

        super().__init__(self.log_queue)

    def manual_flush(self, wait: int = 0):
        """flushes buffer

        Parameters
        ----------
        wait : int, optional
            seconds to wait if wait > 0, by default 0
        """

        if wait > 0:
            time.sleep(wait)

        with LOCK:
            self.log_handler.flush()

    def prepare(self, record: LogRecord) -> LogRecord:
        """Pass trough for LogRecord

        Parameters
        ----------
        record : LogRecord

        Returns
        -------
        LogRecord
        """
        return record

    @set_interval(10)
    def timout_flush(self):
        """Runs flush every 10 seconds"""
        self.manual_flush()

    def __exit_flush(self):
        self.queue_listener.stop()
        self.manual_flush()
